
from .gen.ageLexer import ageLexer
from .gen.ageParser import ageParser
from .gen.ageVisitor import ageVisitor
from .models import *
from antlr4 import *

class ResultHandler:
    def handleRow(rawString):
        pass


def buildGraph(cursor, resultHandler:ResultHandler=None):
    if resultHandler == None:
        resultHandler = Antlr4ResultHandler(cursor.query)
   
    graph = Graph(cursor.query)

    for record in cursor:
        parsed = resultHandler.handleRow(record[0])
        graph.append(parsed)

    return graph

def getRows(cursor):
    resultHandler = Antlr4ResultHandler(cursor.query)
    for record in cursor:
        yield resultHandler.handleRow(record[0])

class Antlr4ResultHandler(ResultHandler):
    def __init__(self, query=None):
        self.lexer = ageLexer()
        self.parser = ageParser(None)
        self.visitor = ResultVisitor()
        self.graph = Graph(query)

    def handleRow(self, rawString):
        self.lexer.inputStream = InputStream(rawString)
        self.parser.setTokenStream(CommonTokenStream(self.lexer))
        self.parser.reset()
        tree = self.parser.ageout()
        parsed = tree.accept(self.visitor)
        return parsed

    def getGraph(self):
        return self.graph

# print raw result String
class DummyResultHandler(ResultHandler):
    def handleRow(self, rawString):
        print(rawString)

# default ageout visitor
class ResultVisitor(ageVisitor):
    def visitAgeout(self, ctx:ageParser.AgeoutContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#vertex.
    def visitVertex(self, ctx:ageParser.VertexContext):
        vertex = Vertex()
        proCtx = ctx.getTypedRuleContext(ageParser.PropertiesContext,0)

        dict = proCtx.accept(self)
        vertex.id = dict["id"]
        vertex.label = dict["label"]
        vertex.properties = dict["properties"]
        
        return vertex


    # Visit a parse tree produced by ageParser#edge.
    def visitEdge(self, ctx:ageParser.EdgeContext):
        edge = Edge()
        proCtx = ctx.getTypedRuleContext(ageParser.PropertiesContext,0)

        dict = proCtx.accept(self)
        edge.id = dict["id"]
        edge.label = dict["label"]
        edge.end_id = dict["end_id"]
        edge.start_id = dict["start_id"]
        edge.properties = dict["properties"]
        
        return edge


    # Visit a parse tree produced by ageParser#path.
    def visitPath(self, ctx:ageParser.PathContext):
        path = Path()

        vctx1 = ctx.getTypedRuleContext(ageParser.VertexContext, 0)
        ectx = ctx.getTypedRuleContext(ageParser.EdgeContext, 0)
        vctx2 = ctx.getTypedRuleContext(ageParser.VertexContext, 1)

        path.start = vctx1.accept(self)
        path.rel = ectx.accept(self)
        path.end = vctx2.accept(self)
        
        return path


    # Visit a parse tree produced by ageParser#value.
    def visitValue(self, ctx:ageParser.ValueContext):
        c = ctx.getChild(0)
        if isinstance(c, ageParser.PropertiesContext) or isinstance(c,ageParser.ArrContext):
            val = c.accept(self)
            return val
        else:
            val = c.getText().strip('"')
            return val



    # Visit a parse tree produced by ageParser#properties.
    def visitProperties(self, ctx:ageParser.PropertiesContext):
        props = dict()
        for c in ctx.getChildren():
            if isinstance(c, ageParser.PairContext):
                namVal = self.visitPair(c)
                name = namVal[0]
                valCtx = namVal[1]
                val = valCtx.accept(self) 
                props[name] = val
        return props

    # Visit a parse tree produced by ageParser#pair.
    def visitPair(self, ctx:ageParser.PairContext):
        self.visitChildren(ctx)
        return (ctx.STRING().getText().strip('"') , ctx.value())


    # Visit a parse tree produced by ageParser#arr.
    def visitArr(self, ctx:ageParser.ArrContext):
        li = list()
        for c in ctx.getChildren():
            if not isinstance(c, TerminalNode):
                val = c.accept(self)
                li.append(val)
        return li




import psycopg2 
from psycopg2 import errors
from .exceptions import *
from .builder import buildGraph, getRows, ResultHandler 


_EXCEPTION_NoConnection = NoConnection()
_EXCEPTION_GraphNotSet = GraphNotSet()
_EXCEPTION_NoCursor = NoCursor()


class Age:
    def __init__(self):
        self.conn = None    # psycopg2 connection
        self.cursor = None  # psycopg2 cursor
        self.graphName = None

    def connect(self, dsn=None, graph=None, connection_factory=None, cursor_factory=None, **kwargs):
        conn = psycopg2.connect(dsn, connection_factory, cursor_factory, **kwargs)
        cursor = conn.cursor()
        self.conn = conn
        self.cursor = cursor
        self.cursor.execute("LOAD 'age';")
        self.cursor.execute("SET search_path = ag_catalog, '$user', public;")
        if graph != None:
            try :
                self.setGraph(graph)
            except GraphNotFound:
                self.createGraph(graph)
                self.setGraph(graph)

        return self

    def close(self):
        self.cursor.close()
        self.conn.close()

    def setGraph(self, graph):
        if self.graphExists(graph) :
            # apache age doesn't have configuration parameter "graph_path" 
            # self.cursor.execute("SET graph_path = %s", (graph,))
            self.graphName = graph
            return self
        else:
            raise GraphNotFound(graph)
        return self

    def graphExists(self, graph):
        self.cursor.execute("SELECT count(*) FROM ag_graph WHERE name=%s", (graph,))
        return self.cursor.fetchone()[0] > 0 

    def createGraph(self, graphName):
        if self.graphExists(graphName) :
            raise GraphAlreadyExists(graphName)
        else:
            self.cursor.execute("SELECT create_graph(%s);", (graphName,))

    def deleteGraph(self, graphName):
        if self.graphExists(graphName) :
            self.cursor.execute("SELECT drop_graph(%s, true);", (graphName,))
            

    def newCursor(self):
        self.cursor.close()
        self.cursor = self.conn.cursor()
        return self

    def _checkReady(self):
        if self.graphName == None:
            raise _EXCEPTION_GraphNotSet
        if self.conn == None:
            raise _EXCEPTION_NoConnection
        if self.cursor == None:
            raise _EXCEPTION_NoCursor

    def _checkGraphName(self):
        if self.graphName == None:
            raise _EXCEPTION_GraphNotSet
            

    def commit(self):
        self.conn.commit()
        
    def rollback(self):
        self.conn.rollback()
        
    def execSql(self, stmt, *args):
        self._checkReady()
        try:
            self.cursor.execute(stmt, *args)
            return self
        except Exception as cause:
            self.conn.rollback()
            raise SqlExcutionError(stmt,cause)
        
    
    def execCypher(self, cypherStmt, *args):
        self._checkGraphName()
        stmt = "SELECT * from cypher('"+self.graphName+"', $$ "+cypherStmt+" $$) as (v agtype);"
        return self.execSql(stmt, *args)

    def querySql(self, stmt, *args):
        return self.execSql(stmt, *args)
        

    def queryCypher(self, cypherStmt, *args):
        return self.execCypher(cypherStmt, *args)
    
    def graph(self, resultHandler:ResultHandler=None):
        return buildGraph(self.cursor, resultHandler)

    def rows(self):
        return getRows(self.cursor)
    


# Generated from age.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .ageParser import ageParser
else:
    from ageParser import ageParser

# This class defines a complete generic visitor for a parse tree produced by ageParser.

class ageVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by ageParser#ageout.
    def visitAgeout(self, ctx:ageParser.AgeoutContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#vertex.
    def visitVertex(self, ctx:ageParser.VertexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#edge.
    def visitEdge(self, ctx:ageParser.EdgeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#path.
    def visitPath(self, ctx:ageParser.PathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#value.
    def visitValue(self, ctx:ageParser.ValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#properties.
    def visitProperties(self, ctx:ageParser.PropertiesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#pair.
    def visitPair(self, ctx:ageParser.PairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by ageParser#arr.
    def visitArr(self, ctx:ageParser.ArrContext):
        return self.visitChildren(ctx)



del ageParser


import json
from io import StringIO 


TP_NONE = 0
TP_VERTEX = 1
TP_EDGE = 2
TP_PATH = 3


class Graph():
    def __init__(self, stmt=None) -> None:
        self.statement = stmt
        self.rows = list()

    def __iter__(self):
        return self.rows.__iter__()

    def __len__(self):
        return self.rows.__len__()

    def append(self, row):
        self.rows.append(row)
        
class AGObj:
    @property
    def gtype(self):
        return TP_NONE


class Path(AGObj):
    def __init__(self, start=None, rel=None, end=None) -> None:
        self.start = start
        self.rel = rel
        self.end = end

    @property
    def gtype(self):
        return TP_PATH

    def __str__(self) -> str:
        return self.toString()

    def __repr__(self) -> str:
        return self.toString()

    def toString(self) -> str: 
        buf = StringIO()
        buf.write("[")
        if self.start != None:
            self.start._toString(buf)
   
        buf.write(",")
        if self.rel != None:
            self.rel._toString(buf)
            
        buf.write(",")

        if self.end != None:
            self.end._toString(buf)

        buf.write("]::PATH")

        return buf.getvalue()

    def toJson(self) -> str:
        buf = StringIO()
        buf.write("{\"gtype\": \"path\", ")
        
        buf.write("\"start\": ")
        if self.start != None:
            self.start._toJson(buf)

        buf.write(", \"rel\": ")
        if self.rel != None:
            self.rel._toJson(buf)
            
        buf.write(", \"end\": ")

        if self.end != None:
            self.end._toJson(buf)

        buf.write("}")

        return buf.getvalue()


    

class Vertex(AGObj):
    def __init__(self, id=None, label=None, properties=None) -> None:
        self.id = id
        self.label = label
        self.properties = properties

    @property
    def gtype(self):
        return TP_VERTEX

    def __str__(self) -> str:
        return self.toString()

    def __repr__(self) -> str:
        return self.toString()

    def toString(self) -> str: 
        return nodeToString(self)

    def _toString(self, buf): 
        _nodeToString(self, buf)

    def toJson(self) -> str:
        return nodeToJson(self)

    def _toJson(self, buf): 
        _nodeToJson(self, buf)


class Edge(AGObj):
    def __init__(self, id=None, label=None, properties=None) -> None:
        self.id = id
        self.label = label
        self.start_id = None
        self.end_id = None
        self.properties = properties

    @property
    def gtype(self):
        return TP_EDGE

    def __str__(self) -> str:
        return self.toString()

    def __repr__(self) -> str:
        return self.toString()

    def extraStrFormat(node, buf):
        if node.start_id != None:
            buf.write(", start_id:")
            buf.write(node.start_id)

        if node.end_id != None:
            buf.write(", end_id:")
            buf.write(node.end_id)


    def toString(self) -> str: 
        return nodeToString(self, Edge.extraStrFormat)

    def _toString(self, buf): 
        _nodeToString(self, buf, Edge.extraStrFormat)

    def extraJsonFormat(node, buf):
        if node.start_id != None:
            buf.write(", \"start_id\": \"")
            buf.write(node.start_id)
            buf.write("\"")

        if node.end_id != None:
            buf.write(", \"end_id\": \"")
            buf.write(node.end_id)
            buf.write("\"")

    def toJson(self) -> str:
        return nodeToJson(self, Edge.extraJsonFormat)

    def _toJson(self, buf): 
        _nodeToJson(self, buf, Edge.extraJsonFormat)


def nodeToString(node, extraFormatter=None):
    buf = StringIO() 
    _nodeToString(node,buf,extraFormatter=extraFormatter)
    return buf.getvalue()


def _nodeToString(node, buf, extraFormatter=None):
    buf.write("{")
    if node.label != None:
        buf.write("label:")
        buf.write(node.label)
        
    if node.id != None:
        buf.write(", id:")
        buf.write(node.id)
        
    if node.properties != None:
        buf.write(", properties:{")
        for k,v in node.properties.items():
            buf.write(k)
            buf.write(": ")
            buf.write(v)
            buf.write(",")
        buf.write("}")

    if extraFormatter != None:
        extraFormatter(node, buf)
        
    if node.gtype == TP_VERTEX:
        buf.write("}::VERTEX")
    if node.gtype == TP_EDGE:
        buf.write("}::EDGE")


def nodeToJson(node, extraFormatter=None):
    buf = StringIO()
    _nodeToJson(node, buf, extraFormatter=extraFormatter)
    return buf.getvalue()


def _nodeToJson(node, buf, extraFormatter=None):
    buf.write("{\"gtype\": ")
    if node.gtype == TP_VERTEX:
        buf.write("\"vertex\", ")
    if node.gtype == TP_EDGE:
        buf.write("\"edge\", ")

    if node.label != None:
        buf.write("\"label\":\"")
        buf.write(node.label)
        buf.write("\"")
        
    if node.id != None:
        buf.write(", \"id\":")
        buf.write(node.id)
        
    if extraFormatter != None:
        extraFormatter(node, buf)
        
    if node.properties != None:
        buf.write(", \"properties\":{")
        for k,v in node.properties.items():
            buf.write("\"")
            buf.write(k)
            buf.write("\": \"")
            buf.write(v)
            buf.write("\",")
        buf.write("}")
    buf.write("}")
    


from . import age
from .age import *
from .models import *
from .builder import ResultHandler, DummyResultHandler, buildGraph, getRows
from . import VERSION 

def version():
    return VERSION.VERSION


def connect(dsn=None, graph=None, connection_factory=None, cursor_factory=None, **kwargs):
        ag = Age()
        ag.connect(dsn=dsn, graph=graph, connection_factory=connection_factory, cursor_factory=cursor_factory, **kwargs)
        return ag

# Dummy ResultHandler
rawPrinter = DummyResultHandler()

__name__="age"


import unittest
import age 

DSN = "host=172.17.0.3 port=5432 dbname=postgres user=postgres password=agens"

class TestAgeBasic(unittest.TestCase):
    ag = None

    def __init__(self, methodName: str) -> None:
        super().__init__(methodName=methodName)
        self.ag = age.connect(dsn=DSN, graph="test_graph")

    def __del__(self):
        # Clear test data
        self.ag.execCypher("MATCH (n:Person) DETACH DELETE n RETURN *")
        self.ag.execSql("DROP TABLE IF EXISTS infos")
        self.ag.commit()
        self.ag.deleteGraph(self.ag.graphName)
        self.ag.close()
    
    def tearDown(self) -> None:
        self.ag.execCypher("MATCH (n:Person) DETACH DELETE n RETURN *")
        self.ag.commit()
        return super().tearDown()

    def test_vertices(self):
        ag = self.ag
        # Create Vertices
        ag.execCypher("CREATE (n:Person {name: 'Joe'})")
        ag.execCypher("CREATE (n:Person {name: 'Smith'})")
        ag.execCypher("CREATE (n:Person {name: %s})", ('Jack',))
        ag.execCypher("CREATE (n:Person {name: 'Andy', title: 'Developer'})")
        ag.execCypher("CREATE (n:Person {name: %s, title: %s})", ('Tom','Developer',))
        ag.commit()

        # Query Vertices with result parsed full graph.
        ag.queryCypher("MATCH (n:Person) RETURN n")
        graph = ag.graph()
        self.assertEqual(5, len(graph), "Create and retrieve vertices.")

        for vertex in graph:
            self.assertEqual( age.models.Vertex, type(vertex), "Retrieve and unmarshal Vertex.")

        # Query Vertices with parsed row cursor.
        ag.queryCypher("MATCH (n:Person) RETURN n")
        for vertex in ag.rows():
            self.assertIsNotNone(vertex.id)
            self.assertIsNotNone(vertex.properties["name"])
            self.assertIsNotNone(vertex.toJson())

    def test_paths(self):
        ag = self.ag
        # Create Vertices
        ag.execCypher("CREATE (n:Person {name: 'Joe'})")
        ag.execCypher("CREATE (n:Person {name: 'Smith'})")
        ag.execCypher("CREATE (n:Person {name: %s})", ('Jack',))
        ag.execCypher("CREATE (n:Person {name: 'Andy', title: 'Developer'})")
        ag.execCypher("CREATE (n:Person {name: %s, title: %s})", ('Tom','Developer',))
        ag.commit()
        # Create Edges
        ag.execCypher("MATCH (a:Person), (b:Person) WHERE a.name = 'Joe' AND b.name = 'Smith' CREATE (a)-[r:workWith {weight: 3}]->(b)")
        ag.execCypher("MATCH (a:Person), (b:Person) WHERE  a.name = 'Andy' AND b.name = 'Tom' CREATE (a)-[r:workWith {weight: 1}]->(b)")
        ag.execCypher("MATCH (a:Person {name: 'Jack'}), (b:Person {name: 'Andy'}) CREATE (a)-[r:workWith {weight: 5}]->(b)")
        ag.execCypher("MATCH (a:Person {name: 'Joe'}), (b:Person {name: 'Jack'}) CREATE (a)-[r:workWith {weight: 5}]->(b)")
        ag.commit()

        ag.queryCypher("MATCH p=()-[:workWith]-() RETURN p")
        graph = ag.graph()
        self.assertEqual(8, len(graph), "Create and retrieve paths.")
        for path in graph:
            self.assertEqual( age.models.Path, type(path), "Retrieve and unmarshal Path.")
            

if __name__ == '__main__':
    unittest.main()


# Generated from age.g4 by ANTLR 4.9.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\21")
        buf.write("S\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7\4\b")
        buf.write("\t\b\4\t\t\t\3\2\3\2\3\2\3\2\5\2\27\n\2\3\3\3\3\3\3\3")
        buf.write("\4\3\4\3\4\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\6\3\6")
        buf.write("\3\6\3\6\3\6\3\6\3\6\5\6/\n\6\3\7\3\7\3\7\3\7\7\7\65\n")
        buf.write("\7\f\7\16\78\13\7\3\7\3\7\3\7\3\7\5\7>\n\7\3\b\3\b\3\b")
        buf.write("\3\b\3\t\3\t\3\t\3\t\7\tH\n\t\f\t\16\tK\13\t\3\t\3\t\3")
        buf.write("\t\3\t\5\tQ\n\t\3\t\2\2\n\2\4\6\b\n\f\16\20\2\2\2W\2\26")
        buf.write("\3\2\2\2\4\30\3\2\2\2\6\33\3\2\2\2\b\36\3\2\2\2\n.\3\2")
        buf.write("\2\2\f=\3\2\2\2\16?\3\2\2\2\20P\3\2\2\2\22\27\5\n\6\2")
        buf.write("\23\27\5\4\3\2\24\27\5\6\4\2\25\27\5\b\5\2\26\22\3\2\2")
        buf.write("\2\26\23\3\2\2\2\26\24\3\2\2\2\26\25\3\2\2\2\27\3\3\2")
        buf.write("\2\2\30\31\5\f\7\2\31\32\7\f\2\2\32\5\3\2\2\2\33\34\5")
        buf.write("\f\7\2\34\35\7\r\2\2\35\7\3\2\2\2\36\37\7\3\2\2\37 \5")
        buf.write("\4\3\2 !\7\4\2\2!\"\5\6\4\2\"#\7\4\2\2#$\5\4\3\2$%\7\5")
        buf.write("\2\2%&\7\16\2\2&\t\3\2\2\2\'/\7\17\2\2(/\7\20\2\2)/\5")
        buf.write("\f\7\2*/\5\20\t\2+/\7\6\2\2,/\7\7\2\2-/\7\b\2\2.\'\3\2")
        buf.write("\2\2.(\3\2\2\2.)\3\2\2\2.*\3\2\2\2.+\3\2\2\2.,\3\2\2\2")
        buf.write(".-\3\2\2\2/\13\3\2\2\2\60\61\7\t\2\2\61\66\5\16\b\2\62")
        buf.write("\63\7\4\2\2\63\65\5\16\b\2\64\62\3\2\2\2\658\3\2\2\2\66")
        buf.write("\64\3\2\2\2\66\67\3\2\2\2\679\3\2\2\28\66\3\2\2\29:\7")
        buf.write("\n\2\2:>\3\2\2\2;<\7\t\2\2<>\7\n\2\2=\60\3\2\2\2=;\3\2")
        buf.write("\2\2>\r\3\2\2\2?@\7\17\2\2@A\7\13\2\2AB\5\n\6\2B\17\3")
        buf.write("\2\2\2CD\7\3\2\2DI\5\n\6\2EF\7\4\2\2FH\5\n\6\2GE\3\2\2")
        buf.write("\2HK\3\2\2\2IG\3\2\2\2IJ\3\2\2\2JL\3\2\2\2KI\3\2\2\2L")
        buf.write("M\7\5\2\2MQ\3\2\2\2NO\7\3\2\2OQ\7\5\2\2PC\3\2\2\2PN\3")
        buf.write("\2\2\2Q\21\3\2\2\2\b\26.\66=IP")
        return buf.getvalue()


class ageParser ( Parser ):

    grammarFileName = "age.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'['", "','", "']'", "'true'", "'false'", 
                     "'null'", "'{'", "'}'", "':'", "'::vertex'", "'::edge'", 
                     "'::path'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "KW_VERTEX", "KW_EDGE", 
                      "KW_PATH", "STRING", "NUMBER", "WS" ]

    RULE_ageout = 0
    RULE_vertex = 1
    RULE_edge = 2
    RULE_path = 3
    RULE_value = 4
    RULE_properties = 5
    RULE_pair = 6
    RULE_arr = 7

    ruleNames =  [ "ageout", "vertex", "edge", "path", "value", "properties", 
                   "pair", "arr" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    KW_VERTEX=10
    KW_EDGE=11
    KW_PATH=12
    STRING=13
    NUMBER=14
    WS=15

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.9.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class AgeoutContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self):
            return self.getTypedRuleContext(ageParser.ValueContext,0)


        def vertex(self):
            return self.getTypedRuleContext(ageParser.VertexContext,0)


        def edge(self):
            return self.getTypedRuleContext(ageParser.EdgeContext,0)


        def path(self):
            return self.getTypedRuleContext(ageParser.PathContext,0)


        def getRuleIndex(self):
            return ageParser.RULE_ageout

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAgeout" ):
                listener.enterAgeout(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAgeout" ):
                listener.exitAgeout(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAgeout" ):
                return visitor.visitAgeout(self)
            else:
                return visitor.visitChildren(self)




    def ageout(self):

        localctx = ageParser.AgeoutContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_ageout)
        try:
            self.state = 20
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 16
                self.value()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 17
                self.vertex()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 18
                self.edge()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 19
                self.path()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VertexContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def properties(self):
            return self.getTypedRuleContext(ageParser.PropertiesContext,0)


        def KW_VERTEX(self):
            return self.getToken(ageParser.KW_VERTEX, 0)

        def getRuleIndex(self):
            return ageParser.RULE_vertex

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVertex" ):
                listener.enterVertex(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVertex" ):
                listener.exitVertex(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVertex" ):
                return visitor.visitVertex(self)
            else:
                return visitor.visitChildren(self)




    def vertex(self):

        localctx = ageParser.VertexContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_vertex)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 22
            self.properties()
            self.state = 23
            self.match(ageParser.KW_VERTEX)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EdgeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def properties(self):
            return self.getTypedRuleContext(ageParser.PropertiesContext,0)


        def KW_EDGE(self):
            return self.getToken(ageParser.KW_EDGE, 0)

        def getRuleIndex(self):
            return ageParser.RULE_edge

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEdge" ):
                listener.enterEdge(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEdge" ):
                listener.exitEdge(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitEdge" ):
                return visitor.visitEdge(self)
            else:
                return visitor.visitChildren(self)




    def edge(self):

        localctx = ageParser.EdgeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_edge)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 25
            self.properties()
            self.state = 26
            self.match(ageParser.KW_EDGE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PathContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def vertex(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ageParser.VertexContext)
            else:
                return self.getTypedRuleContext(ageParser.VertexContext,i)


        def edge(self):
            return self.getTypedRuleContext(ageParser.EdgeContext,0)


        def KW_PATH(self):
            return self.getToken(ageParser.KW_PATH, 0)

        def getRuleIndex(self):
            return ageParser.RULE_path

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPath" ):
                listener.enterPath(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPath" ):
                listener.exitPath(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPath" ):
                return visitor.visitPath(self)
            else:
                return visitor.visitChildren(self)




    def path(self):

        localctx = ageParser.PathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_path)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self.match(ageParser.T__0)
            self.state = 29
            self.vertex()
            self.state = 30
            self.match(ageParser.T__1)
            self.state = 31
            self.edge()
            self.state = 32
            self.match(ageParser.T__1)
            self.state = 33
            self.vertex()
            self.state = 34
            self.match(ageParser.T__2)
            self.state = 35
            self.match(ageParser.KW_PATH)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(ageParser.STRING, 0)

        def NUMBER(self):
            return self.getToken(ageParser.NUMBER, 0)

        def properties(self):
            return self.getTypedRuleContext(ageParser.PropertiesContext,0)


        def arr(self):
            return self.getTypedRuleContext(ageParser.ArrContext,0)


        def getRuleIndex(self):
            return ageParser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValue" ):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)




    def value(self):

        localctx = ageParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_value)
        try:
            self.state = 44
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [ageParser.STRING]:
                self.enterOuterAlt(localctx, 1)
                self.state = 37
                self.match(ageParser.STRING)
                pass
            elif token in [ageParser.NUMBER]:
                self.enterOuterAlt(localctx, 2)
                self.state = 38
                self.match(ageParser.NUMBER)
                pass
            elif token in [ageParser.T__6]:
                self.enterOuterAlt(localctx, 3)
                self.state = 39
                self.properties()
                pass
            elif token in [ageParser.T__0]:
                self.enterOuterAlt(localctx, 4)
                self.state = 40
                self.arr()
                pass
            elif token in [ageParser.T__3]:
                self.enterOuterAlt(localctx, 5)
                self.state = 41
                self.match(ageParser.T__3)
                pass
            elif token in [ageParser.T__4]:
                self.enterOuterAlt(localctx, 6)
                self.state = 42
                self.match(ageParser.T__4)
                pass
            elif token in [ageParser.T__5]:
                self.enterOuterAlt(localctx, 7)
                self.state = 43
                self.match(ageParser.T__5)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropertiesContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pair(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ageParser.PairContext)
            else:
                return self.getTypedRuleContext(ageParser.PairContext,i)


        def getRuleIndex(self):
            return ageParser.RULE_properties

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProperties" ):
                listener.enterProperties(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProperties" ):
                listener.exitProperties(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitProperties" ):
                return visitor.visitProperties(self)
            else:
                return visitor.visitChildren(self)




    def properties(self):

        localctx = ageParser.PropertiesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_properties)
        self._la = 0 # Token type
        try:
            self.state = 59
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 46
                self.match(ageParser.T__6)
                self.state = 47
                self.pair()
                self.state = 52
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==ageParser.T__1:
                    self.state = 48
                    self.match(ageParser.T__1)
                    self.state = 49
                    self.pair()
                    self.state = 54
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 55
                self.match(ageParser.T__7)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 57
                self.match(ageParser.T__6)
                self.state = 58
                self.match(ageParser.T__7)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PairContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(ageParser.STRING, 0)

        def value(self):
            return self.getTypedRuleContext(ageParser.ValueContext,0)


        def getRuleIndex(self):
            return ageParser.RULE_pair

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPair" ):
                listener.enterPair(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPair" ):
                listener.exitPair(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPair" ):
                return visitor.visitPair(self)
            else:
                return visitor.visitChildren(self)




    def pair(self):

        localctx = ageParser.PairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_pair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 61
            self.match(ageParser.STRING)
            self.state = 62
            self.match(ageParser.T__8)
            self.state = 63
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ageParser.ValueContext)
            else:
                return self.getTypedRuleContext(ageParser.ValueContext,i)


        def getRuleIndex(self):
            return ageParser.RULE_arr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArr" ):
                listener.enterArr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArr" ):
                listener.exitArr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArr" ):
                return visitor.visitArr(self)
            else:
                return visitor.visitChildren(self)




    def arr(self):

        localctx = ageParser.ArrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_arr)
        self._la = 0 # Token type
        try:
            self.state = 78
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,5,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 65
                self.match(ageParser.T__0)
                self.state = 66
                self.value()
                self.state = 71
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==ageParser.T__1:
                    self.state = 67
                    self.match(ageParser.T__1)
                    self.state = 68
                    self.value()
                    self.state = 73
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 74
                self.match(ageParser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 76
                self.match(ageParser.T__0)
                self.state = 77
                self.match(ageParser.T__2)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx









from psycopg2.errors import *

class GraphNotFound(Exception):
    def __init__(self, name):
        self.name = name

    def __repr__(self) :
        return 'Graph[' + self.name + '] does not exist.' 


class GraphAlreadyExists(Exception):
    def __init__(self, name):
        self.name = name

    def __repr__(self) :
        return 'Graph[' + self.name + '] already exists.' 

        
class GraphNotSet(Exception):
    def __repr__(self) :
        return 'Graph name is not set.'


class NoConnection(Exception):
    def __repr__(self) :
        return 'No Connection'

class NoCursor(Exception):
    def __repr__(self) :
        return 'No Cursor'

class SqlExcutionError(Exception):
    def __init__(self, msg, cause):
        self.msg = msg
        self.cause = cause
        super().__init__(msg, cause)
    
    def __repr__(self) :
        return 'SqlExcution [' + self.msg + ']'  



# Generated from age.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .ageParser import ageParser
else:
    from ageParser import ageParser

# This class defines a complete listener for a parse tree produced by ageParser.
class ageListener(ParseTreeListener):

    # Enter a parse tree produced by ageParser#ageout.
    def enterAgeout(self, ctx:ageParser.AgeoutContext):
        pass

    # Exit a parse tree produced by ageParser#ageout.
    def exitAgeout(self, ctx:ageParser.AgeoutContext):
        pass


    # Enter a parse tree produced by ageParser#vertex.
    def enterVertex(self, ctx:ageParser.VertexContext):
        pass

    # Exit a parse tree produced by ageParser#vertex.
    def exitVertex(self, ctx:ageParser.VertexContext):
        pass


    # Enter a parse tree produced by ageParser#edge.
    def enterEdge(self, ctx:ageParser.EdgeContext):
        pass

    # Exit a parse tree produced by ageParser#edge.
    def exitEdge(self, ctx:ageParser.EdgeContext):
        pass


    # Enter a parse tree produced by ageParser#path.
    def enterPath(self, ctx:ageParser.PathContext):
        pass

    # Exit a parse tree produced by ageParser#path.
    def exitPath(self, ctx:ageParser.PathContext):
        pass


    # Enter a parse tree produced by ageParser#value.
    def enterValue(self, ctx:ageParser.ValueContext):
        pass

    # Exit a parse tree produced by ageParser#value.
    def exitValue(self, ctx:ageParser.ValueContext):
        pass


    # Enter a parse tree produced by ageParser#properties.
    def enterProperties(self, ctx:ageParser.PropertiesContext):
        pass

    # Exit a parse tree produced by ageParser#properties.
    def exitProperties(self, ctx:ageParser.PropertiesContext):
        pass


    # Enter a parse tree produced by ageParser#pair.
    def enterPair(self, ctx:ageParser.PairContext):
        pass

    # Exit a parse tree produced by ageParser#pair.
    def exitPair(self, ctx:ageParser.PairContext):
        pass


    # Enter a parse tree produced by ageParser#arr.
    def enterArr(self, ctx:ageParser.ArrContext):
        pass

    # Exit a parse tree produced by ageParser#arr.
    def exitArr(self, ctx:ageParser.ArrContext):
        pass



del ageParser


from setuptools import setup, find_packages
from age import VERSION 

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

setup(
    name             = 'apache-age-python',
    version          = VERSION.VERSION,
    description      = 'Python driver for Apache AGE',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author           = 'Ikchan Kwon',
    author_email     = 'rhizome.ai@gmail.com',
    url              = 'https://github.com/rhizome-ai/apache-age-python',
    download_url     = 'https://github.com/rhizome-ai/apache-age-python/releases/tag/v'+VERSION.VERSION ,
    license          = 'Apache2.0',
    install_requires = [ 'psycopg2', 'antlr4-python3-runtime' ],
    packages         = find_packages(exclude = ['bisect', 'notebooks', 'samples']),
    keywords         = ['Graph Database', 'Apache AGE', 'PostgreSQL'],
    python_requires  = '>=3.9',
    # package_data     =  {},
    # zip_safe=False,
    classifiers      = [
        'Programming Language :: Python :: 3.9'
    ]
)